#Be sure that SEQRES rows are recorded in the PDB files.
#Command line usage: The working directory is where the script runs.
#e.g. --> 'python getNtDisSum.py'

#!/usr/bin/env python

import os
import re
import csv

class CoordDict(object):	#CoordDict class can be used to store nucleotides' 3-dimensional coordinates and calculate Euclidean distances between any two of them.
    def __init__(self, ID, coord):	#ID --> PDB ID; coord --> Coordinates dictionary.
        self.ID = ID
        self.coord = coord

    def getDis(self, chain1='A', chain2='A', nt1=1, nt2=2):
        self.dim = len(self.coord[chain1][nt1])
        try:
            tmp = 0
            for i in xrange(self.dim):
                tmp += pow(self.coord[chain1][nt1][i] - self.coord[chain2][nt2][i], 2)
            dis = pow(tmp, 0.5)
            return dis
        except IndexError, diag:
            print diag

class PDB(object):	#PDB class can be used to load PDB files and parse the information about nucleotide in them.
    def __init__(self, fileName):	#fileName --> PDB file name.
        self.fName = fileName
        self.ID = fileName.split('.')[0]

    def getSeqDict(self):
        self.seqDict = {}	#keys --> chain IDs; values --> sequence strings.
        self.ntDict = {}	#keys --> chain IDs; values --> lists of nucleotides in each sequence.
        self.chainID = []	#A list of chain IDs.
        self.chainType = dict()	#keys --> chain IDs; values --> type of each chain.
        typeDict = dict(zip([1,2,3],['RNA','DNA','Peptide']))

        pattern = re.compile(r'(?P<field>^SEQRES)\s*(?P<serNum>\d+)\s*(?P<chainID>\S+)\s*(?P<numNt>\d+)\s*(?P<ntName>.*)')
        #A pattern used to match the 'SEQRES' rows.

        with open(self.fName, 'r') as f:
            for eachLine in f:
                m = pattern.search(eachLine)
                if (m is not None) and ((len(self.chainID) == 0) or ((len(self.chainID) > 0) and (m.group('chainID') != self.chainID[-1]))):
                    self.chainID.append(m.group('chainID'))
                    self.seqDict[self.chainID[-1]] = (m.group('ntName').strip())
                elif (m is not None) and (len(self.chainID) > 0) and (m.group('chainID') == self.chainID[-1]):	#The same chain.
                    self.seqDict[self.chainID[-1]] += (' '+m.group('ntName').strip())
                elif (m is None) and (self.seqDict == {}):	#Before reaching the 'SEQRES' rows.
                    continue
                elif (m is None) and (self.seqDict != {}):	#Finished the 'SEQRES' rows.
                    self.chainNum = len(self.chainID)
                    print '%r has %d chain(s).\n' % (self.ID, self.chainNum)
                    for key in self.chainID:
                        self.seqDict[key] = re.sub(r' {2,3}',r' ',self.seqDict[key])	#Replace the continuous spaces into single ones.
                        self.ntDict[key] = self.seqDict[key].split()
                        self.chainType[key] = typeDict[min([len(nt) for nt in self.ntDict[key]])]
                        print 'The length of chain%r(%s) is %d:\n%r\n' % (key, self.chainType[key], len(self.ntDict[key]), self.seqDict[key])
                    f.seek(0)
                    break
        return self.seqDict

    def getNtCoordDict(self, atmName="C1'"):	#Default atom used to calculate distances between nucleotides is set to C1'.
        self.getSeqDict()

        self.ntSeq = {}		#keys --> chain IDs; values --> lists of nucleotides' sequence IDs.
        self.ntName = {}	#keys --> chain IDs; values --> lists of nucleotides' names.
        self.coord = {}		#keys --> chain IDs; values --> lists of nucleotides' coordinates.

        pattern = re.compile(r'(?P<field>^ATOM|^HETATM)\s*(?P<serNum>\d+)\s(?P<atmName>.{4})(?P<altLoc>.)(?P<ntName>.{3})\s(?P<chainID>.)(?P<ntSeq>.{5})\s*(?P<x>-?\d+\.\d+)\s*(?P<y>-?\d+\.\d+)\s*(?P<z>-?\d+\.\d+)\s*(?P<occupancy>-?\d+\.\d+)\s*(?P<tempFactor>-?\d+\.\d+)\s*(?P<element>\S+)(?P<charge>.{2})\s*')
        #A pattern used to match the 'ATOM' and 'HETATM' rows.

        print 'Processing the nucleotide atom(%s) coords(xyz) of %r...\n' % (atmName, self.ID)
        for key in self.chainID:
            if self.chainType[key] != 'RNA':
                continue
            self.ntSeq[key] = []
            self.ntName[key] = []
            self.coord[key] = []
            x, y, z = 0, 0, 0
            flag = 0
            with open(self.fName, 'r') as f:
                for eachLine in f:
                    m = pattern.search(eachLine)
                    if (m is not None) and (m.group('chainID') == key):
                        if (m.group('atmName').strip() == atmName) and ((m.group('altLoc') == ' ') or (m.group('altLoc') == 'A')):	#altLoc --> alternative locations. If 'altLoc' has more than one value, then we only consider the first('A').
                            x, y, z = float(m.group('x')), float(m.group('y')), float(m.group('z'))
                            self.coord[key].append((x, y, z))
                            self.ntSeq[key].append(m.group('ntSeq').strip().lower())
                            self.ntName[key].append(m.group('ntName').strip())
                            flag = 1
                    elif ('ANISOU' in eachLine):	#Skip the 'ANISOU' rows.
                        continue
                    elif (m is None) and (flag == 0):	#Before reaching the 'ATOM' and 'HETATM' rows.
                        continue
                    elif (m is None) and (flag == 1):	#Finished the 'ATOM' and 'HETATM' rows.
                        break
                print 'Successfully got %d nucleotide atom(%s) coords(xyz) on chain%r of %r.\n' % (len(self.ntName[key]), atmName, key, self.ID)
        return self.coord

def getNtDisDict():	#Calculate all the Euclidean distances between any two nucleotides and store the data in a dictionary.
    global atmName, flist, pdbid, pdbf, pdb, ncd
    atmName = "C1'"	#Default atom used to calculate distances between nucleotides is set to C1'.
    flist = os.listdir(os.getcwd())
    print 'There is/are %d file(s) in your current working directory.' % (len(flist))
    pdbid = []
    pdbf = []
    for each in flist:
        if '.pdb' in each:
            pdbid.append(each.split('.')[0])
            pdbf.append(each)
            print '\t', each
    print 'There is/are %d PDB file(s) in your current working directory.' % (len(pdbid))

    pdb, ncd = {}, {}	#keys --> PDB IDs; values --> instances of PDB/CoordDict class.
    ntDisDict = {}	#keys --> 'PDB ID_chain ID1_chain ID2's; values --> lists of lists containing the distances between two corresponding nucleotides.

    for i in pdbid:
        pdb[i] = PDB(pdbf[pdbid.index(i)])
        ncd[i] = CoordDict(i, pdb[i].getNtCoordDict(atmName))
        for m in pdb[i].chainID:
            ntCoordNumm = len(ncd[i].coord[m])
            if ntCoordNumm:	#If the chain m is amino acid chain, then ntCoordNumm would be 0(False).
                for n in pdb[i].chainID:
                    ntCoordNumn = len(ncd[i].coord[n])
                    if ntCoordNumn:	#If the chain n is amino acid chain, then ntCoordNumn would be 0(False).
                        ntDisDict[i+'_'+m+'_'+n] = []
                        for j in xrange(ntCoordNumm):
                            ntDisDict[i+'_'+m+'_'+n].append([])
                            for k in xrange(ntCoordNumn):
                                ntDisDict[i+'_'+m+'_'+n][j].append(ncd[i].getDis(m, n, j, k))
    return ntDisDict

def getNtDisSum(ntDisDict):	#For each nucleotide, sum up the distances between the nucleotide and all the others and use a dictionary to save them.
    ntDisSum = {}	#keys --> 'PDB ID_chain ID's; values --> lists of distance sum between each nucleotide and the rest.
    ntKey = {}		#keys --> PDB IDs; values --> lists of 'chain ID_ntName ntSequence ID's.
    for ID in pdbid:
        ndsName = ID+'_atom('+atmName+')_nds.txt'
        if (ndsName not in flist):
            ntDisSum[ID] = []
            ntKey[ID] = []
            for x in pdb[ID].chainID:
                ntDisSum[ID+'_'+x] = []
                for key in sorted(ntDisDict.keys()):
                    if ((ID == key.split('_')[0]) and (x == key.split('_')[1])):
                        ntCoordNum = len(ntDisDict[key])
                        if (len(ntDisSum[ID+'_'+x]) < ntCoordNum):	#Calculate on chain x for the first time.
                            for i in xrange(ntCoordNum):
                                ntDisSum[ID+'_'+x].append(sum(ntDisDict[key][i]))
                        elif (len(ntDisSum[ID+'_'+x]) == ntCoordNum):	#Calculate on chain x with another chain.
                            for i in xrange(ntCoordNum):
                                ntDisSum[ID+'_'+x][i] += sum(ntDisDict[key][i])
                ntDisSum[ID] += ntDisSum[ID+'_'+x]	#Concatenate the chains.
                chainLen = len(ntDisSum[ID+'_'+x])	#Length of chain.
                for i in xrange(chainLen):
                    ntKey[ID].append(x+'_'+pdb[ID].ntName[x][i]+pdb[ID].ntSeq[x][i])
            fields=['#','nt','nds'];ntDisSumRows = []	#Contents to write into the '*_nds.txt' files.
            molLen = len(ntDisSum[ID])	#Length of molecule.
            for i in xrange(molLen):
                ntDisSumRows.append(dict(zip(fields,[i+1,ntKey[ID][i],round(ntDisSum[ID][i],2)])))
            try:
                with open(ndsName, 'w') as f:
                    writer=csv.DictWriter(f,fieldnames=fields,dialect='tab')
                    writer.writerow(dict(zip(fields,fields)))
                    writer.writerows(ntDisSumRows)
                    print 'Successfully saved the results to %r in your current working directory.\n' % (ndsName)
            except IOError, diag:
                print diag
    return ntDisSum

def main():
    csv.register_dialect('tab',delimiter='\t',quoting=csv.QUOTE_NONNUMERIC,lineterminator='\n')
    ndd = getNtDisDict()
    nds = getNtDisSum(ndd)

if (__name__ == '__main__'):
    main()
